import org.neo4j.graphdb.GraphDatabaseService;
import org.neo4j.io.fs.FileUtils;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;

/**
 * Created by hanngu on 05.05.15.
 */
public class Algorithm {

    private int numberOfIterations, antColonySize;
    private double percentOfCrazyAnts, inertiaWeight, pheromoneVisited, pheromoneBest, evaporation, percentOfAntsToBeFollowed, percentOfAntsThatAreGettingMorePheromoneBasedOnEvaluation;
    private GenerateNetwork generateNetwork;
    private Statistics statistics;
    int numberOfIterationsThatDidNotSatisfyInitialDemand = 0;

    public Algorithm(int numberOfIterations, double percentOfCrazyAnts, double inertiaWeight,
                     double pheromoneVisited, double pheromoneBest, double evaporation, int antColonySize,
                     double percentOfAntsToBeFollowed, double percentOfAntsThatAreGettingMorePheromoneBasedOnEvaluation, GenerateNetwork generateNetwork) {
        this.numberOfIterations = numberOfIterations;
        this.percentOfCrazyAnts = percentOfCrazyAnts;
        this.inertiaWeight = inertiaWeight;
        this.pheromoneVisited = pheromoneVisited;
        this.pheromoneBest = pheromoneBest;
        this.evaporation = evaporation;
        this.antColonySize = antColonySize;
        this.percentOfAntsToBeFollowed = percentOfAntsToBeFollowed;
        this.percentOfAntsThatAreGettingMorePheromoneBasedOnEvaluation = percentOfAntsThatAreGettingMorePheromoneBasedOnEvaluation;
        this.generateNetwork = generateNetwork;
        statistics = new Statistics();

    }

    public Ant runAlgorithm(Neo4jHandler neo4jHandler) {
        neo4jHandler.createNeo4jDatabase(generateNetwork);
        ArrayList<MandlNodeCouple> mandlNodeCouples = generateNetwork.generateNodeCouples(neo4jHandler);
        double totalDemandForNetwork = generateNetwork.calculateTotalDemandForNetwork(mandlNodeCouples);
        ArrayList<MandlNode> mandlNodes = generateNetwork.getMandlNodes();
        ArrayList<MandlEdge> mandlEdges = generateNetwork.getMandlEdges();


        //Creates antColony
        SuperSwarm aco = new SuperSwarm(neo4jHandler, statistics, numberOfIterations, inertiaWeight, percentOfCrazyAnts, pheromoneVisited, pheromoneBest, percentOfAntsToBeFollowed, percentOfAntsThatAreGettingMorePheromoneBasedOnEvaluation, totalDemandForNetwork);
        //StandardACO aco = new StandardACO(neo4jHandler, statistics, numberOfIterations, pheromoneVisited, pheromoneBest, totalDemandForNetwork);
        aco.setNodeCouples(mandlNodeCouples);
        int maxNodesInRoute = 8;
        int minNodesInRoute = 2;
        for (int i = 0; i < numberOfIterations; i++) {
            ArrayList<Ant> ants = aco.generateAnts(antColonySize, mandlNodes, maxNodesInRoute, minNodesInRoute);

            //Generate routes for all ants
            aco.createRoutes(4, ants, mandlEdges);

            //Evaluate ants
            aco.evaluation(ants);
            if(aco.isNoAntsSatisfiedInitialDemand()){
                numberOfIterationsThatDidNotSatisfyInitialDemand ++;
                continue;
            }

            //Evaporate pheromones
            aco.evaporation(evaporation);

        }
        Ant bestAnt = aco.getBestAnt();
        //BestAntStatistics bestAntStatistics = new BestAntStatistics(statistics.round(bestAnt.getD0(), 2), statistics.round(bestAnt.calculateAverageTravelTime(mandlNodeCouples.size(), totalDemandForNetwork), 2));
        return bestAnt;
    }

    public int getNumberOfIterationsThatDidNotSatisfyInitialDemand() {
        return numberOfIterationsThatDidNotSatisfyInitialDemand;
    }
}
