import java.util.ArrayList;

/**
 * Created by hanngu on 18.02.15.
 */
public class Ant {

    private ArrayList<ArrayList<Integer>> routeSet;
    private int antNumber, maxNodesInRoute, minNodesInRoute;
    private ArrayList<Integer> currentRoute;
    private MandlNode startMandlNode, currentMandlNode;
    private double f1, f2, f3, d0, d1, d2, dUnsat;
    private double totalTravelTime;
    private double demandSatisfiedByDirectTravel;
    private double demandSatisfiedByOneTransfer;
    private double demandSatisfiedByTwoTransfers;
    private double demandNotSatisfied;
    private double averageTravelTime;
    private boolean createdRouteWithLessThanMinNodes;

    private ArrayList<MandlNodeCouple> directCouples, oneTransferCouples, twoTransferCouples, unsatisfiedCouples;

    private ArrayList<MandlEdge> edgesInRouteSet;
    private boolean crazyAnt, isFollower;

    public Ant(int antNumber, MandlNode startMandlNode, int maxNodesInRoute, boolean crazyAnt,
               int minNodesInRoute){
        this.antNumber = antNumber;
        this.startMandlNode = startMandlNode;
        this.maxNodesInRoute = maxNodesInRoute;
        this.minNodesInRoute = minNodesInRoute;
        this.crazyAnt = crazyAnt;
        this.isFollower = false;
        this.createdRouteWithLessThanMinNodes = false;
        currentMandlNode = startMandlNode;
        routeSet = new ArrayList<>();
        currentRoute = new ArrayList<>();
        edgesInRouteSet = new ArrayList<>();
        this.f1 = 0.0;
        this.f2 = 0.0;
        this.f3 = 0.0;
        this.d0 = 0.0;
        this.d1 = 0.0;
        this.d2 = 0.0;
        this.dUnsat = 0.0;
        this.totalTravelTime = 0.0;
        this.averageTravelTime = -1;
        this.directCouples = new ArrayList<>();
        this.oneTransferCouples = new ArrayList<>();
        this.twoTransferCouples = new ArrayList<>();
        this.unsatisfiedCouples = new ArrayList<>();
        currentRoute.add(startMandlNode.getNumber());
    }

    public Ant(Ant a, int antNumber, boolean crazyAnt) {
        this.antNumber = antNumber;
        this.crazyAnt = crazyAnt;
        this.startMandlNode = a.getStartMandlNode();
        this.maxNodesInRoute = a.getMaxNodesInRoute();
        this.isFollower = true;
        currentMandlNode = a.getCurrentMandlNode();
        this.routeSet = a.copyRouteSet(a);
        this.currentRoute = a.copyCurrentRoute(a);
        this.edgesInRouteSet = copyEdgesInRouteSet(a);
        this.f1 = a.getF1();
        this.f2 = a.getF2();
        this.f3 = a.getF3();
        this.d0 = a.getD0();
        this.d1 = a.getD1();
        this.d2 = a.getD2();
        this.dUnsat = a.getdUnsat();
        this.totalTravelTime = a.getTotalTravelTime();
        this.directCouples = a.getDirectCouples();
        this.oneTransferCouples = a.getOneTransferCouples();
        this.twoTransferCouples = a.getTwoTransferCouples();
        this.unsatisfiedCouples = a.getUnsatisfiedCouples();
    }

    public Ant(ArrayList<ArrayList<Integer>> routeSet){
        this.routeSet = routeSet;
        this.totalTravelTime = 0.0;
        this.antNumber = 1;
    }

    public boolean createdRouteWithLessThanMinNodes() {
        return createdRouteWithLessThanMinNodes;
    }

    public double getAverageTravelTime() {
        if(averageTravelTime == -1){
            System.err.println("Average travel time is not sat for this ant");
            return -1;
        }
        return averageTravelTime;
    }

    public boolean isCrazyAnt() {
        return crazyAnt;
    }

    public MandlNode getStartMandlNode() {
        return startMandlNode;
    }

    private ArrayList<ArrayList<Integer>> copyRouteSet(Ant a) {
        ArrayList<ArrayList<Integer>> routeSetToReturn = new ArrayList<>();
        for(ArrayList<Integer> route : a.getRouteSet()){
            ArrayList<Integer> copyOfRoute = copyCurrentRoute(route);
            routeSetToReturn.add(copyOfRoute);
        }
        return routeSetToReturn;
    }

    private ArrayList<Integer> copyCurrentRoute(ArrayList<Integer> route) {
        ArrayList<Integer> routeToReturn = new ArrayList<>();
        for(Integer i : route){
            routeToReturn.add(i);
        }
        return routeToReturn;
    }

    private ArrayList<Integer> copyCurrentRoute(Ant a) {
        ArrayList<Integer> routeToReturn = new ArrayList<>();
        for(Integer i : a.getCurrentRoute()){
            routeToReturn.add(i);
        }
        return routeToReturn;
    }

    private ArrayList<MandlEdge> copyEdgesInRouteSet(Ant a) {
        ArrayList<MandlEdge> edgesToReturn = new ArrayList<>();
        for(MandlEdge edge : a.edgesInRouteSet){
            edgesToReturn.add(edge);
        }
        return edgesToReturn;
    }

    public int getMaxNodesInRoute() {
        return maxNodesInRoute;
    }

    public ArrayList<MandlEdge> getEdgesInRouteSet() {
        return edgesInRouteSet;
    }

    public void addEdge(MandlEdge edge){
        edgesInRouteSet.add(edge);
    }

    public double getD0() {
        return d0;
    }

    public void setD0(double d0) {
        this.d0 = d0;
    }

    public double getD1() {
        return d1;
    }

    public void setD1(double d1) {
        this.d1 = d1;
    }

    public double getD2() {
        return d2;
    }

    public void setD2(double d2) {
        this.d2 = d2;
    }

    public double getdUnsat() {
        return dUnsat;
    }

    public void setdUnsat(double dUnsat) {
        if(dUnsat > 0.001){
            this.dUnsat = dUnsat;
        } else {
            this.dUnsat = 0.0;
        }
    }

    public double calculateAverageTravelTime(int nodeCouplesSize, double totalNetworkDemand){
        this.averageTravelTime = totalTravelTime/totalNetworkDemand;
        return averageTravelTime;
    }

    public double getTotalTravelTime() {
        return totalTravelTime;
    }

    public void setTotalTravelTime(double averageTravelTime) {
        this.totalTravelTime = averageTravelTime;
    }

    public int getAntNumber() {
        return antNumber;
    }

    //f1 is the dif between shortest path i the network and the shortest path possible given the route set of ant a for every nodeCouple in nodeCouples. This number should be as small as possible
    public void setF1(double f1) {
        this.f1 = f1;
    }

    public double getF1() {
        return f1;
    }

    public double getF2() {
        return f2;
    }

    public double getF3() {
        return f3;
    }

    //A small f2 value indicates an ant with many direct routes
    public void setF2(double f2) {
        this.f2 = f2;
    }

    //f3 reflects how many unsatisfied nodeCouples that excists, and should very small (0)
    public void setF3(double f3) {
        this.f3 = f3;
    }

    public double getTotaltFit(){
        return f1+f2+f3;
    }

    public ArrayList<Integer> getCurrentRoute() {
        return currentRoute;
    }

    public void finishedRoute(Neo4jHandler neo4jHandler, int iterationNumber){
        createNewRoute(neo4jHandler, iterationNumber);
    }

    public boolean setNextNode(MandlNode mandlNode, Neo4jHandler neo4jHandler, int iterationNumber){
        //Must follow the constraints - no backtracking or cycles allowed.
        if(currentRoute.size() < maxNodesInRoute && !currentRoute.contains(mandlNode.getNumber())) {
            currentRoute.add(mandlNode.getNumber());
            currentMandlNode = mandlNode;
            currentMandlNode.setAttractiveValue(currentMandlNode.getAttractiveValue()/10);
            return true;

        } else{
            createNewRoute(neo4jHandler, iterationNumber);
            return false;
        }
    }

    private void createNewRoute(Neo4jHandler neo4jHandler, int iterationNumber){
        if(currentRoute.size() < minNodesInRoute){
            createdRouteWithLessThanMinNodes = true;
        }
        routeSet.add(new ArrayList<Integer>(currentRoute));
        neo4jHandler.addRouteToGraph(currentRoute, routeSet.size(), antNumber, iterationNumber);
        currentRoute = new ArrayList<Integer>();
        //currentMandlNode = AntMain.generateRandomNode();
        //currentRoute.add(currentMandlNode.getNumber());
    }

    public MandlNode getCurrentMandlNode() {
        return currentMandlNode;
    }

    public ArrayList<ArrayList<Integer>> getRouteSet() {
        return routeSet;
    }

    public String toString(){
        String temp = "";
        temp += "Routes: [";
        for(ArrayList<Integer> route : routeSet){
            temp += "{";
            for(Integer number : route){
                temp += number+",";
            }
            temp += "}";
        }
        temp += "]";

        return "number: " + antNumber + " startMandlNode: " + startMandlNode.getNumber() + " currentMandlNode: " + currentMandlNode.getNumber() + temp;
    }

    public double getDemandSatisfiedByOneTransfer() {
        return this.demandSatisfiedByOneTransfer;
    }

    public double getDemandSatisfiedByDirectTravel() {
        return demandSatisfiedByDirectTravel;
    }

    public double getDemandSatisfiedByTwoTransfers() {
        return this.demandSatisfiedByTwoTransfers;

    }
    public void setDemandSatisfiedByTwoTransfers(double demandSatisfiedByTwoTransfers) {
        this.demandSatisfiedByTwoTransfers = demandSatisfiedByTwoTransfers;
    }

    public void setDemandSatisfiedByOneTransfer(double demandSatisfiedByOneTransfer) {
        this.demandSatisfiedByOneTransfer = demandSatisfiedByOneTransfer;
    }

    public void setDemandSatisfiedByDirectTravel(double demandSatisfiedByDirectTravel) {
        this.demandSatisfiedByDirectTravel = demandSatisfiedByDirectTravel;
    }

    public void setDemandNotSatisfied(double demandNotSatisfied) {
        this.demandNotSatisfied = demandNotSatisfied;
    }

    public ArrayList<MandlNodeCouple> getDirectCouples() {
        return directCouples;
    }

    public void setDirectCouples(ArrayList<MandlNodeCouple> directCouples) {
        this.directCouples = directCouples;
    }

    public ArrayList<MandlNodeCouple> getOneTransferCouples() {
        return oneTransferCouples;
    }

    public void setOneTransferCouples(ArrayList<MandlNodeCouple> oneTransferCouples) {
        this.oneTransferCouples = oneTransferCouples;
    }

    public ArrayList<MandlNodeCouple> getTwoTransferCouples() {
        return twoTransferCouples;
    }

    public void setTwoTransferCouples(ArrayList<MandlNodeCouple> twoTransferCouples) {
        this.twoTransferCouples = twoTransferCouples;
    }

    public ArrayList<MandlNodeCouple> getUnsatisfiedCouples() {
        return unsatisfiedCouples;
    }

    public void setUnsatisfiedCouples(ArrayList<MandlNodeCouple> unsatisfiedCouples) {
        this.unsatisfiedCouples = unsatisfiedCouples;
    }

    public boolean isFollower() {
        return isFollower;
    }
}
