import org.neo4j.io.fs.FileUtils;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;

/**
 * Created by hanngu on 09.05.15.
 */
public class MainWithOutLoops {

    public static void main(String[] args) {

        int numberOfIterations = 125;
        double percentOfCrazyAnts = 0.25; //percent
        double inertiaWeight = 1.0;
        double pheromoneVisited = 0.1;
        double pheromoneBest = 1.3;
        double evaporation = 0.25; //percent to be removed
        int antColonySize = 50;
        double percentOfAntsToBeFollowed = 0.05;
        double percentOfAntsThatAreGettingMorePheromoneBasedOnEvaluation = percentOfAntsToBeFollowed;


        GenerateNetwork generateNetwork = new GenerateNetwork();

        //Creates Nodes
        generateNetwork.createNodes("CEC2013SuppCLM (1)/Instances/MandlCoords.txt");

        //Creates Edges
        generateNetwork.generateEdges("CEC2013SuppCLM (1)/Instances/MandlDemand.txt");


        //Creates Demands
        generateNetwork.generateTravelTimes("CEC2013SuppCLM (1)/Instances/MandlTravelTimes.txt");

        //Creates a list of connecting edges in all mandlNodes
        generateNetwork.generateListOfEdgesInNodes();

        //Calculates total demand in each node
        generateNetwork.calculateTotalDemandForEachNode("CEC2013SuppCLM (1)/Instances/MandlDemand.txt");

        Algorithm algorithm = new Algorithm(numberOfIterations, percentOfCrazyAnts, inertiaWeight,
                pheromoneVisited, pheromoneBest, evaporation, antColonySize, percentOfAntsToBeFollowed,
                percentOfAntsThatAreGettingMorePheromoneBasedOnEvaluation,  generateNetwork);


        String filename = "Neo4jTest" + args[0];
        try {
            FileUtils.deleteRecursively(new File(filename));
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            Neo4jHandler neo4jHandler = new Neo4jHandler(filename);
            long startTime = System.nanoTime();
            Ant bestAnt = algorithm.runAlgorithm(neo4jHandler);
            long endTime = System.nanoTime();
            long duration = (endTime-startTime)/1000000;
            int second = (int) (duration/1000);
            ArrayList<MandlNodeCouple> nodeCouples = generateNetwork.getMandlNodeCouples();
            double demand = generateNetwork.calculateTotalDemandForNetwork(nodeCouples);
            System.out.println("*");
            System.out.println("Iteration number:" + args[0]);
            System.out.println("Seconds:" + second );
            System.out.println("Best Ant Totfit:" + bestAnt.getTotaltFit());
            System.out.println("Best Ant D0:" + bestAnt.getD0());
            System.out.println("Best Ant D1:" + bestAnt.getD1());
            System.out.println("Best Ant D2:" + bestAnt.getD2());
            System.out.println("Best Ant dUnsat:" + bestAnt.getdUnsat());
            System.out.println("Best Ant ATT:" + bestAnt.calculateAverageTravelTime(nodeCouples.size(), demand));
            String temp = "Best Ants Route: {";
            for(ArrayList<Integer> route : bestAnt.getRouteSet()){
                temp += "[";
                for(Integer i : route){
                    temp += i+",";
                }
                temp += "]";
            }
            temp+="}";
            //System.out.println("Percentage of iterations where no ants satisfied the initial constraint: " + ((algorithm.getNumberOfIterationsThatDidNotSatisfyInitialDemand() / (double) numberOfIterations) * 100));
            System.out.println(temp);
            System.out.println();
            neo4jHandler.shutdown();
            neo4jHandler.cleanUp();
        }

    }
}
