import java.util.ArrayList;
import java.util.Collections;
import java.util.Random;

/**
 * Created by hanngu on 09.05.15.
 */
public class StandardACO extends ACO{
    private boolean noAntsSatisfiedInitialDemand;

    public StandardACO(Neo4jHandler neo4jHandler, Statistics statistics, int numberOfIterations, double pheromoneBest, double pheromoneVisited
            , double totalDemand) {
        super(neo4jHandler, statistics, numberOfIterations, pheromoneVisited, pheromoneBest, totalDemand);
        noAntsSatisfiedInitialDemand = false;
    }

    @Override
    public void evaluation(ArrayList<Ant> ants) {
        ArrayList<Ant> bestAnts = new ArrayList<>();
        ArrayList<Integer> nodeNumbers = new ArrayList<>();
        for (MandlNode n : nodes) nodeNumbers.add(n.getNumber());
        for (Ant a : ants) {
            if(a.isFollower()){
                bestAnts.add(a);
                continue;
            }
            if(a.getTotaltFit() != 0){
                continue;
            }
            ArrayList<Integer> copyOfNodeNumbers = new ArrayList<>();
            for (Integer i : nodeNumbers)
                copyOfNodeNumbers.add(i);

            for (ArrayList<Integer> route : a.getRouteSet()) {
                if (!copyOfNodeNumbers.isEmpty()) {
                    for (Integer n : nodeNumbers) {
                        if (route.contains(n) && copyOfNodeNumbers.contains(n)) {
                            copyOfNodeNumbers.remove(n);
                        }
                    }
                }
            }
            if (copyOfNodeNumbers.isEmpty()) {
                bestAnts.add(a);
            }
        }
        //Generates couple of mandlNodes, that should be used in evaluation. The number of node1 is at [0] in the array, and the number of node2 is at [1] in the array
        //ArrayList<MandlNodeCouple> totalNodeCouples = createNodeCouples(mandlNodes, neo4jHandler);
        ArrayList<Ant> disconnectedAnts = new ArrayList<>();
        for (Ant a : bestAnts) {
            if(a.isFollower()){
                continue;
            }
            //Calculates F1
            //Calculates the dif between shortest path i the network and the shortest path possible given the route set of ant a for every nodeCouple in nodeCouples. This number should be as small as possible
            double dif = 0;

            dif += neo4jHandler.findShortestPathForCurrentAnt(a, nodeCouples, iterationNumber);

            if (dif == -1) {
                disconnectedAnts.add(a);
                continue;
            }
            dif = dif/Math.pow(nodes.size(), 2);
            a.setF1(dif);
            //Calculates F2
            //Calculates a score based on number of direct travelers, travelers with only one transfer and travelers with only two transfers. The returned value is the number of MandlNodeCouples that fulfills these criterias
            double satisfiedCouples = calculateF2(a, nodeCouples);

            //Calculates F3
            //Adds the number of unsatisfied MandleNodeCouples (totalNodeCouples-satisfiedCouples) to the ant. A good ant has as few as possible of these.
            double unsatisfied = (nodeCouples.size() - satisfiedCouples)*10;
            a.setF3(unsatisfied);
        }
        //The java 8 way
        disconnectedAnts.forEach(bestAnts::remove);
        if(bestAnts.size() == 0){
            noAntsSatisfiedInitialDemand = true;
        }
        if(!noAntsSatisfiedInitialDemand){
            Collections.sort(bestAnts, new TotfitComparator());
            if (bestAnt == null || bestAnt.getTotaltFit() > bestAnts.get(0).getTotaltFit()) {
                bestAnt = bestAnts.get(0);
            }
        }

    }

    @Override
    protected MandlEdge selectEdge(ArrayList<MandlEdge> possibleMandlEdges, Ant a) {

        ArrayList<MandlEdge> edgesWithUnexploredNodes = new ArrayList<>();
        for (MandlEdge e : possibleMandlEdges) {
            if (e.getMandlNode1().getNumber() == a.getCurrentMandlNode().getNumber()) {
                if (!checkIfNodeIsInAntsRouteSet(a, e.getMandlNode2())) {
                    edgesWithUnexploredNodes.add(e);
                }
            } else {
                if (!checkIfNodeIsInAntsRouteSet(a, e.getMandlNode1())) {
                    edgesWithUnexploredNodes.add(e);
                }
            }
        }

        double totalPheromone = 0.0;


        for (MandlEdge e : possibleMandlEdges) {
            totalPheromone += e.getPheromone();
        }
        if (totalPheromone == 0) {
            totalPheromone = -1;
        }


        double totalEdgeValue = 0;

        for (MandlEdge e : possibleMandlEdges) {
            double edgeValue = Math.abs((e.getPheromone() / totalPheromone));
            totalEdgeValue += edgeValue;
            e.setEdgeValue(edgeValue);

        }
        if (totalEdgeValue == 0) {
            double tempValue = 1.0;
            boolean firstIter = true;
            for (MandlEdge e : possibleMandlEdges) {
                if (!firstIter) {
                    tempValue -= 1 / possibleMandlEdges.size();
                }
                if(e.getProbability() == 0.0){
                    e.setProbability(tempValue);
                }
                totalEdgeValue += 1 / possibleMandlEdges.size();
                firstIter = false;
            }
        } else {
            Collections.sort(possibleMandlEdges, new EdgeValueComparator());
            double range = 1; //Always 1, because the addition of all the probabilities = 1
            boolean first = true;
            for (MandlEdge e : possibleMandlEdges) {
                double probability = e.getEdgeValue()/totalEdgeValue;
                if(first){
                    e.setProbability(range);
                } else {
                    e.setProbability(range-probability);
                }
                range -= probability;

            }

        }
        Random r = new Random();
        double randis = r.nextDouble();
        MandlEdge chosenMandlEdge = null;
        Collections.reverse(possibleMandlEdges);
        for (MandlEdge e : possibleMandlEdges) {
            if (randis < e.getProbability()) {
                chosenMandlEdge = e;
                break;
            }
        }

        if(chosenMandlEdge.isCopy()){
            for(MandlEdge edge : possibleMandlEdges){
                if(edge.isCopy() == false && (chosenMandlEdge.getMandlNode1() == edge.getMandlNode1() && chosenMandlEdge.getMandlNode2() == edge.getMandlNode2())){
                    chosenMandlEdge = edge;
                    break;
                }
            }
        }

        for (MandlEdge e : possibleMandlEdges) {
            e.setProbability(0.0);
            e.setEdgeValue(0.0);
        }

        return chosenMandlEdge;
    }

    public boolean isNoAntsSatisfiedInitialDemand() {
        return noAntsSatisfiedInitialDemand;
    }

}
