import java.util.ArrayList;
import java.util.Collections;
import java.util.Random;

/**
 * Created by hanngu on 16.04.15.
 */
public class SuperSwarm extends ACO{

    private ArrayList<Ant> bestAntsToBeFollowed;
    private double inertiaWeight;
    private double percentOfCrazyAnts;
    private double percentOfAntsToBeFollowed;
    private double percentOfAntsThatAreGettingMorePheromone;
    private boolean noAntsSatisfiedInitialDemand;

    public SuperSwarm(Neo4jHandler neo4jHandler, Statistics statistics, int numberOfIterations, double inertiaWeight, double percentOfCrazyAnts, double pheromoneVisited, double pheromoneBest, double percentOfAntsToBeFollowed, double percentOfAntsThatAreGettingMorePheromone, double totalDemand) {
        super(neo4jHandler, statistics, numberOfIterations, pheromoneVisited, pheromoneBest, totalDemand);
        bestAntsToBeFollowed = new ArrayList<>();
        this.inertiaWeight = inertiaWeight;
        this.percentOfCrazyAnts = percentOfCrazyAnts;
        this.percentOfAntsToBeFollowed = percentOfAntsToBeFollowed;
        this.percentOfAntsThatAreGettingMorePheromone = percentOfAntsThatAreGettingMorePheromone;
        noAntsSatisfiedInitialDemand = false;
    }

   @Override
   public ArrayList<Ant> generateAnts(int colonySize, ArrayList<MandlNode> mandlNodes, int maxNodesInRoute, int minNodesInRoute) {
        iterationNumber ++;

        this.nodes = mandlNodes;
        ArrayList<Ant> ants = new ArrayList<>();
        if(!bestAntsToBeFollowed.isEmpty()) {
            for (int i = 0; i < bestAntsToBeFollowed.size(); i++) {
                boolean crazyAnt = false;
                Ant follower = new Ant(bestAntsToBeFollowed.get(i), i, crazyAnt);
                ants.add(follower);
                walkTheSamePathAsAGoodAnt(follower);
            }
        }
        double probability = percentOfCrazyAnts*inertiaWeight;
        for (int i = bestAntsToBeFollowed.size(); i < colonySize; i++) {
            boolean crazyAnt = false;
            double chance = new Random().nextDouble();
            if(chance <= probability){
                crazyAnt = true;
        }
            MandlNode firstNode = chooseFirstNode();
            Ant ant = new Ant(i, firstNode, maxNodesInRoute, crazyAnt, minNodesInRoute);
            ants.add(ant);
        }
        inertiaWeight = inertiaWeight - (inertiaWeight/numberOfIterations);
        return ants;
    }

    private void walkTheSamePathAsAGoodAnt(Ant follower) {
        for(MandlEdge edge : follower.getEdgesInRouteSet()){
            edge.increasePheromone(pheromoneVisited);
        }
    }


    @Override
    public void evaluation(ArrayList<Ant> ants) {
        noAntsSatisfiedInitialDemand = false;
        bestAntsToBeFollowed = new ArrayList<>();
        ArrayList<Ant> bestAnts = new ArrayList<>();
        ArrayList<Integer> nodeNumbers = new ArrayList<>();
        for (MandlNode n : nodes) nodeNumbers.add(n.getNumber());
        for (Ant a : ants) {

            if(a.isFollower()){
                bestAnts.add(a);
                continue;
            }
            if(a.getTotaltFit() != 0){
                continue;
            }
            if(a.createdRouteWithLessThanMinNodes()){
                continue;
            }
            ArrayList<Integer> copyOfNodeNumbers = new ArrayList<>();
            for (Integer i : nodeNumbers)
                copyOfNodeNumbers.add(i);

            for (ArrayList<Integer> route : a.getRouteSet()) {
                if (!copyOfNodeNumbers.isEmpty()) {
                    for (Integer n : nodeNumbers) {
                        if (route.contains(n) && copyOfNodeNumbers.contains(n)) {
                            copyOfNodeNumbers.remove(n);
                        }
                    }
                }
            }
            if (copyOfNodeNumbers.isEmpty()) {
                bestAnts.add(a);
            }
        }
        //Generates couple of mandlNodes, that should be used in evaluation. The number of node1 is at [0] in the array, and the number of node2 is at [1] in the array
        //ArrayList<MandlNodeCouple> totalNodeCouples = createNodeCouples(mandlNodes, neo4jHandler);
        ArrayList<Ant> disconnectedAnts = new ArrayList<>();
        for (Ant a : bestAnts) {
            if(a.isFollower()){
                continue;
            }
            //Calculates F1
            //Calculates the dif between shortest path i the network and the shortest path possible given the route set of ant a for every nodeCouple in nodeCouples. This number should be as small as possible

            double dif = neo4jHandler.findShortestPathForCurrentAnt(a, nodeCouples, iterationNumber);
            //System.out.println("f1: " + dif);
            if (dif == -1) {
                disconnectedAnts.add(a);
                continue;
            }
            dif = dif/Math.pow(nodes.size(), 2);
            a.setF1(dif);
            //Calculates F2
            //Calculates a score based on number of direct travelers, travelers with only one transfer and travelers with only two transfers. The returned value is the number of MandlNodeCouples that fulfills these criterias
            double satisfiedCouples = calculateF2(a, nodeCouples);

            //Calculates F3
            //Adds the number of unsatisfied MandleNodeCouples (totalNodeCouples-satisfiedCouples) to the ant. A good ant has as few as possible of these.
            double unsatisfied = (nodeCouples.size() - satisfiedCouples)*10;
            a.setF3(unsatisfied);

        }
        //The java 8 way
        disconnectedAnts.forEach(bestAnts::remove);
        if(bestAnts.size() == 0){
            //System.out.println("No ants satisfied the initial demands");
            noAntsSatisfiedInitialDemand = true;
        }
        if(!noAntsSatisfiedInitialDemand){
            Collections.sort(bestAnts, new TotfitComparator());
            if (bestAnt == null || bestAnt.getTotaltFit() > bestAnts.get(0).getTotaltFit()) {
                bestAnt = bestAnts.get(0);
            }
            //Adding to list to be followed in next iteration #BCO 10% of best ants
            double numberOfAntsToBeFollowed = statistics.round(bestAnts.size()*percentOfAntsToBeFollowed, 0);
            for (int i = 0; i < numberOfAntsToBeFollowed ; i++) {
                bestAntsToBeFollowed.add(bestAnts.get(i));
            }

            double numberOfAntsToBeChosen = statistics.round(bestAnts.size()*percentOfAntsThatAreGettingMorePheromone, 0);

            ArrayList<Ant> antsThatAreGettingMorePheromone = new ArrayList<>();
            for (int i = 0; i < numberOfAntsToBeChosen; i++) {
                antsThatAreGettingMorePheromone.add(bestAnts.get(i));
            }

            ArrayList<MandlEdge> edgesThatAreGettingMorePheromone = findFinalBestAntsEdges(antsThatAreGettingMorePheromone);
            giveBestRouteSetsMorePheromone(edgesThatAreGettingMorePheromone);
        }

    }

    @Override
    protected MandlEdge selectEdge(ArrayList<MandlEdge> possibleMandlEdges, Ant a) {

        if(a.isCrazyAnt()){
            int i = new Random().nextInt(possibleMandlEdges.size());
            return possibleMandlEdges.get(i);
        }
        MandlEdge bestAntsChoice = null;
        if(bestAnt != null) {
            //If the current best ant has been in a's currentNode, the edge chosen by the current best ant should be considered.
            bestAntsChoice = getTheEdgeChoosenByBestAnt(possibleMandlEdges, a);
        }

        ArrayList<MandlEdge> edgesWithUnexploredNodes = new ArrayList<>();
        for (MandlEdge e : possibleMandlEdges) {
            if (e.getMandlNode1().getNumber() == a.getCurrentMandlNode().getNumber()) {
                if (!checkIfNodeIsInAntsRouteSet(a, e.getMandlNode2())) {
                    edgesWithUnexploredNodes.add(e);
                }
            } else {
                if (!checkIfNodeIsInAntsRouteSet(a, e.getMandlNode1())) {
                    edgesWithUnexploredNodes.add(e);
                }
            }
        }

        double totalPheromone = 0.0;
        //double totalUnexploredNodes = 0.0;


        for (MandlEdge e : possibleMandlEdges) {
            totalPheromone += e.getPheromone();
        }
        if (totalPheromone == 0) {
            totalPheromone = -1;
        }

        double totalEdgeValue = 0;

        for (MandlEdge e : possibleMandlEdges) {
            double edgeValue = Math.abs((e.getPheromone() / totalPheromone));
            if(edgesWithUnexploredNodes.size() > 0 && edgesWithUnexploredNodes.contains(e)){
                edgeValue+=1;
            }
            totalEdgeValue += edgeValue;
            e.setEdgeValue(edgeValue);

        }

        double staticTotal = totalEdgeValue;
        if (totalEdgeValue == 0) {
            double tempValue = 1.0;
            boolean firstIter = true;
            for (MandlEdge e : possibleMandlEdges) {
                if (!firstIter) {
                    tempValue -= 1 / possibleMandlEdges.size();
                }
                if(e.getProbability() == 0.0){
                    e.setProbability(tempValue);
                }
                totalEdgeValue += 1 / possibleMandlEdges.size();
                firstIter = false;
            }
        } else {
            Collections.sort(possibleMandlEdges, new EdgeValueComparator());
            double range = 1; //Always 1, because the addition of all the probabilities = 1
            boolean first = true;
            for (MandlEdge e : possibleMandlEdges) {
                double probability = e.getEdgeValue()/totalEdgeValue;
                if(first){
                    e.setProbability(range);
                } else {
                    e.setProbability(range-probability);
                }
                range -= probability;

            }

        }
        Random r = new Random();
        double randis = r.nextDouble();
        MandlEdge chosenMandlEdge = null;
        Collections.reverse(possibleMandlEdges);
        for (MandlEdge e : possibleMandlEdges) {
            if (randis < e.getProbability()) {
                chosenMandlEdge = e;
                break;
            }
        }

        if(chosenMandlEdge.isCopy()){
            for(MandlEdge edge : possibleMandlEdges){
                if(edge.isCopy() == false && (chosenMandlEdge.getMandlNode1() == edge.getMandlNode1() && chosenMandlEdge.getMandlNode2() == edge.getMandlNode2())){
                    chosenMandlEdge = edge;
                    break;
                }
            }
        }

        for (MandlEdge e : possibleMandlEdges) {
            e.setProbability(0.0);
            e.setEdgeValue(0.0);
        }

        return chosenMandlEdge;
    }

    private MandlEdge getTheEdgeChoosenByBestAnt(ArrayList<MandlEdge> possibleMandlEdges, Ant a) {
        MandlEdge edgeToReturn = null;
        for(ArrayList<Integer> route: bestAnt.getRouteSet()){
            for(int i = 0; i < route.size()-1; i++){
                if(route.get(i) == a.getCurrentMandlNode().getNumber()){
                    edgeToReturn = findEdgeUsedByBestAnt(route.get(i), route.get(i+1));
                    return edgeToReturn;
                }
            }
        }


        return edgeToReturn;
    }

    private MandlEdge findEdgeUsedByBestAnt(Integer node, Integer nextNode) {
        MandlEdge edgeToReturn = null;
        for(MandlEdge edge : mandlEdges){
            if(edge.getMandlNode1().getNumber() == node && edge.getMandlNode2().getNumber() == nextNode || edge.getMandlNode2().getNumber() == node && edge.getMandlNode1().getNumber() == nextNode){
                edgeToReturn = edge;
            }
        }


        return edgeToReturn;
    }

    public boolean isNoAntsSatisfiedInitialDemand() {
        return noAntsSatisfiedInitialDemand;
    }
}
