import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

/**
 * Created by hanngu on 13.02.15.
 */
public class GenerateNetwork {

    private ArrayList<MandlNode> mandlNodes = new ArrayList<MandlNode>();
    private ArrayList<MandlEdge> mandlEdges = new ArrayList<MandlEdge>();
    private ArrayList<MandlNodeCouple> mandlNodeCouples = new ArrayList<>();
    private double minTotalTravelTimeForNetwork = 0.0;

    public ArrayList<MandlNode> getMandlNodes() {
        return mandlNodes;
    }

    public ArrayList<MandlEdge> getMandlEdges() {
        return mandlEdges;
    }

    public ArrayList<MandlNodeCouple> generateNodeCouples(Neo4jHandler neo4jHandler){
        ArrayList<MandlNodeCouple> nodeCouples = new ArrayList<MandlNodeCouple>();
        for (int i = 0; i < mandlNodes.size()-1; i++) {
            for (int j = i+1; j < mandlNodes.size(); j++) {
                double minTotalTravelTime = neo4jHandler.findShortesPathBetweenTwoNodes(mandlNodes.get(i), mandlNodes.get(j));
                minTotalTravelTimeForNetwork += minTotalTravelTime;
                int demand = 0;
                for(MandlEdge edge : mandlEdges){
                    if(edge.getMandlNode1() == mandlNodes.get(i) && edge.getMandlNode2() == mandlNodes.get(j) || edge.getMandlNode2() == mandlNodes.get(i) && edge.getMandlNode1() == mandlNodes.get(j) ){
                        demand = edge.getDemand();
                        //minTotalTravelTime = minTotalTravelTime;
                    }
                }
                MandlNodeCouple mandlNodeCouple = new MandlNodeCouple(mandlNodes.get(i), mandlNodes.get(j), minTotalTravelTime, demand);
                nodeCouples.add(mandlNodeCouple);
            }
        }
        this.mandlNodeCouples = nodeCouples;

        return nodeCouples;
    }

    public ArrayList<MandlNodeCouple> getMandlNodeCouples() {
        return mandlNodeCouples;
    }

    public double getMinTotalTravelTimeForNetwork() {
        return minTotalTravelTimeForNetwork;
    }

    public void createNodes(String path){

        ArrayList<MandlNode> mandlNodesToBeCreated = new ArrayList<MandlNode>();
        BufferedReader br = null;

        try {

            String sCurrentLine;
            br = new BufferedReader(new FileReader(path));
            String numberOfNodes = br.readLine(); //is needed. Don't really know why.
            int count = 1;
            while ((sCurrentLine = br.readLine()) != null) {
                MandlNode mandlNode;
                String[] splitted = sCurrentLine.split(" ");
                double x = Double.parseDouble(splitted[0]);
                double y = Double.parseDouble(splitted[1]);
                mandlNode = new MandlNode(count, x, y);
                //System.out.println("{name: \"Node" +mandlNode.getNumber()+"\", x:"+mandlNode.getCordX() + ", y:"+mandlNode.getCordY() + "},");
                mandlNodesToBeCreated.add(mandlNode);
                count ++;
            }

        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try {
                if (br != null)br.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
        this.mandlNodes = mandlNodesToBeCreated;
    }

    public void generateEdges(String path){
        ArrayList<MandlEdge> mandlEdgesToBeCreated = new ArrayList<MandlEdge>();
        BufferedReader br = null;
        try {
            String sCurrentLine;
            br = new BufferedReader(new FileReader(path));
            int count = 0;
            while ((sCurrentLine = br.readLine()) != null) {
                if(sCurrentLine.length() != 0){

                    MandlNode mandlNode1 = mandlNodes.get(count);

                    String[] splitted = sCurrentLine.split(" ");
                    int node2Count = 0;
                    for (int i = 0; i < splitted.length ; i++) {

                        if(splitted[i].equals("a")){ //here we changed the main diagonal from all zeros to all 'a's. This was done to take advantage of the symmetric matrix
                            break;
                        }

                        else if(splitted[i].equals("")){ //needed for special cases, because the input file has different spacing between the entries
                            continue;
                        }

                        else {
                            MandlNode mandlNode2 = mandlNodes.get(node2Count++);
                            MandlEdge mandlEdge = new MandlEdge(mandlNode1, mandlNode2, Integer.parseInt(splitted[i]));
                            mandlEdgesToBeCreated.add(mandlEdge);

                        }


                    }

                    count++;
                }

            }

        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try {
                if (br != null)br.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
        this.mandlEdges = mandlEdgesToBeCreated;
    }


    public void generateTravelTimes(String path){
        BufferedReader br = null;
        try {
            String sCurrentLine;
            br = new BufferedReader(new FileReader(path));
            int count = 0;
            while ((sCurrentLine = br.readLine()) != null) {
                if(sCurrentLine.length() != 0){
                    MandlNode mandlNode1 = mandlNodes.get(count);
                    String[] splitted = sCurrentLine.split(" ");
                    int node2Count = 0;
                    for (int i = 0; i < splitted.length ; i++) {
                        if(splitted[i].equals("Inf")){ //if splitted[i] is "Inf", there is no connection between the mandlNodes, and therefor no edge to be created
                            node2Count ++;
                            continue;
                        } else if(splitted[i].equals("")){ //needed for special cases, because the input file has different spacing between the entries
                            continue;
                        }
                        else if(splitted[i].equals("0")){ //because the matrix is symmetric, we only need the numbers before the zero.
                            break;
                        }
                        else {
                            MandlNode mandlNode2 = mandlNodes.get(node2Count++);
                            for(MandlEdge e : mandlEdges){
                                if(e.getMandlNode1() == mandlNode1 && e.getMandlNode2() == mandlNode2){
                                    e.setTravelTime(Integer.parseInt(splitted[i]));
                                    //System.out.println("{source: \"Node" + mandlNode1.getNumber() +"\", target: \"Node"+mandlNode2.getNumber() + "\", type: \"licensing\"},");
                                }
                            }

                        }

                    }
                    count ++;
                }
            }

        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try {
                if (br != null)br.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }

    }

    public void generateListOfEdgesInNodes() {
        for(MandlNode n : mandlNodes){
            for(MandlEdge e : mandlEdges){
                if((e.getMandlNode1() == n || e.getMandlNode2() == n) && e.getTravelTime()!=-1){
                    n.addEdge(e);
                }

            }
        }

    }

    public void calculateTotalDemandForEachNode(String path) {

        BufferedReader br = null;
        try {
            String sCurrentLine;
            br = new BufferedReader(new FileReader(path));
            int count = 0;
            while ((sCurrentLine = br.readLine()) != null) {
                if(sCurrentLine.length() != 0){
                    MandlNode node = mandlNodes.get(count);
                    String[] splitted = sCurrentLine.split(" ");
                    int totaltDemand = 0;
                    for (int i = 0; i < splitted.length ; i++) {
                        if(splitted[i].equals("a")){
                            continue;
                        }
                        else if(splitted[i].equals("")){ //needed for special cases, because the input file has different spacing between the entries
                            continue;
                        } else {
                            totaltDemand += Integer.parseInt(splitted[i]);
                        }
                    }
                    node.setTotalDemand(totaltDemand);
                    count++;
                }

            }

        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try {
                if (br != null)br.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
    }

    public double calculateTotalDemandForNetwork(ArrayList<MandlNodeCouple> mandlNodeCouples) {
        double demandToReturn = 0.0;
        for(MandlNodeCouple couple : mandlNodeCouples){
            demandToReturn += couple.getDemand();
        }
        return demandToReturn;
    }
}
